(({
  t,
  behaviors,
  Toolshed: ts,
  Iconset,
}) => {
  /**
   * A widget for selecting an icon from an iconset.
   */
  class IconSelector {
    /**
     * Create a new icon selector for the icon selection widget.
     *
     * @param {HTMLElement} iconsetSel
     *   The select form input used to provide the value of the iconset.
     * @param {HTMLElement} iconSel
     *   The select form element that is linked to the icon selection.
     */
    constructor(iconsetSel, iconSel) {
      this.iconsetSel = iconsetSel;
      this.iconSel = iconSel;

      this.onIconClick = this.onIconClick.bind(this);
      this.onCtrlClick = this.onCtrlClick.bind(this);
    }

    /**
     * Initialize the selection widget by creating local values.
     */
    init() {
      if (this.isInit) return;

      this.isInit = true;
      this.iconset = '';
      this.icon = this.iconSel.value;
      this.required = !!this.iconSel.attributes.required;

      this.ctrl = ts.createEl('a', { href: '#', class: 'iconset-selector__button' });

      this.iconSel.style.display = 'none';
      this.iconSel.parentNode.insertBefore(this.ctrl, this.iconSel.nextSibling);

      this.selPane = new ts.AttachedPane(this.ctrl, {
        classes: ['iconset-selector__icons'],
        onRemoveItem: (item) => {
          item.removeEventListener('click', this.onIconClick);
        },
      });
      this.selPane.init();

      // Attach the event listeners for select, interactions.
      if (this.iconsetSel.tagName === 'SELECT') {
        this.onIconsetChange = () => this.setIconset(this.iconsetSel.value);
        this.iconsetSel.addEventListener('change', this.onIconsetChange);
      }

      this.ctrl.addEventListener('click', this.onCtrlClick);

      if (this.iconsetSel.value) {
        this.setIconset(this.iconsetSel.value);
      }
    }

    /**
     * Clean-up any resources and remove event listeners.
     */
    destroy() {
      if (!this.init) return;

      this.ctrl.parentNode.removeChild(this.ctrl);
      this.ctrl.removeEventListener('click', this.onCtrlClick);

      this.iconSel.style.display = '';
      this.selPane.destroy();

      if (this.onIconsetChange) {
        this.iconsetSel.removeEventListener('change', this.onIconsetChange);
      }

      this.init = false;
      delete this.selPane;
      delete this.ctrl;
    }

    /**
     * Change the current icon selected.
     *
     * @param {object|null} icon
     *   Icon data to set as the current icon. If this parameter is empty, then
     *   the icon value and display is interpreted as have no icon selected.
     */
    setIcon(icon = null) {
      if (icon && icon.id) {
        this.icon = icon.id;
        this.iconSel.value = icon.id;
        this.ctrl.innerHTML = icon.html;
      }
      else {
        this.icon = '';
        this.iconSel.value = '';
        this.ctrl.innerHTML = '';
      }
    }

    /**
     * Change the current set of icons to a new set.
     *
     * @param {string} id
     *   Name of the iconset to load as the select widget.
     */
    setIconset(id) {
      // Already on this iconset, no need to update or change anything.
      if (this.iconset === id) return;

      const icons = Iconset.getLoaded(id);

      // Check if
      if (Array.isArray(icons)) {
        this.setIconsetData(id, icons);
      }
      else if (icons === false) {
        this.selPane.close();
        this.iconset = '';
      }
      else {
        this.selPane.clear();

        Iconset.getAsync(id).then((data) => this.setIconsetData(id, data));
      }
    }

    /**
     * Internal function for setting the icons in the dialog wrapper.
     *
     * @param {string} iconsetId
     *   Machine name of the iconset being set.
     * @param {array} data
     *   The icon data as an array of icon info with a `html` property.
     *
     * @return {array}
     *   Return the data passed in unmodified to allow chaining with a Promise
     *   if multiple selectors add then() callbacks when waiting for the loading
     *   to complete.
     */
    setIconsetData(iconsetId, data) {
      if (this.iconset !== iconsetId) {
        this.iconset = iconsetId;
        this.selPane.clear();

        if (data) {
          if (!this.required) {
            // Prepend and empty icon selection when an icon is not required.
            const el = this.createIconSelector({ id: '', label: t('None'), html: '' });
            this.selPane.append(el);
          }

          data.forEach((icon) => {
            const el = this.createIconSelector(icon);
            this.selPane.append(el);
          });
        }

        // Icon does not exist in the new set, clear the value
        if (!this.icon || !data.has(this.icon)) {
          this.setIcon(null);
        }
      }

      return data;
    }

    createIconSelector(icon) {
      const el = ts.createEl('a', { href: '#', class: 'icon-selector' });

      el.innerHTML = icon.html;
      el._iconData = icon;
      el.addEventListener('click', this.onIconClick, true);

      if (icon.id === this.icon) {
        el.classList.add('icon--active');
        this.ctrl.innerHTML = icon.html;
      }

      return el;
    }

    /**
     * Click event for activating the selection pane.
     *
     * @param {MouseEvent} e
     *   The mouse click event information.
     */
    onCtrlClick(e) {
      e.preventDefault();

      if (this.iconset || this.selPane.isVisible()) {
        this.selPane.toggle();
      }
    }

    /**
     * Event handler for when an icon selection is clicked.
     *
     * @param {MouseEvent} e
     *   The mouse click event information.
     */
    onIconClick(e) {
      e.preventDefault();

      const target = e.currentTarget;
      this.setIcon(target._iconData);
      this.selPane.close();

      Array.prototype.forEach.call(
        this.selPane.getContent().getElementsByClassName('icon--active'),
        (el) => el.classList.remove('icon--active'),
      );

      target.classList.add('icon--active');
    }
  }

  /**
   * Create a global tracking for loaded data of iconsets.
   *
   * The map keys are the iconset IDs and the values are data objects provide
   * the information on how the icons are rendered.
   *
   * @type {Map}
   */
  IconSelector.iconsets = new Map();

  /**
   * Icon selector behavior, to find and apply JS to icon selector widgets.
   */
  behaviors.iconsetSelector = {
    selectors: new Map(),

    attach(context) {
      ts.walkByClass(context, 'js-iconset-selector', (el) => {
        const iconSelector = new IconSelector(
          el.getElementsByClassName('icon-sel__iconset')[0],
          el.getElementsByClassName('icon-sel__icon')[0],
        );

        this.selectors.set(el.id, iconSelector);
        iconSelector.init();
      }, 'iconset-selector--processed');
    },
    detach(context, settings, trigger) {
      if (trigger !== 'unload') return;

      ts.walkByClass(context, '.js-iconset-selector', (el) => {
        const iconSel = this.selectors.get(el.id);

        if (iconSel) {
          this.selectors.delete(el.id);
          iconSel.destroy();
        }
      });
    },
  };
})(Drupal);
