<?php

namespace Drupal\iconset;

use Drupal\Component\Plugin\PluginInspectionInterface;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\DependencyInjection\ClassResolverInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Extension\ThemeHandlerInterface;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\Plugin\Discovery\ContainerDerivativeDiscoveryDecorator;
use Drupal\Core\Plugin\Discovery\YamlDiscovery;
use Drupal\Core\Utility\Error;
use Drupal\toolshed\Plugin\StaticYamlPluginManager;

/**
 * Plugin manager for finding and creating iconsets instances.
 */
class IconsetPluginManager extends StaticYamlPluginManager implements IconsetPluginManagerInterface {

  use LoggerChannelTrait;

  /**
   * List of internal and global iconset labels, keyed by the iconset ID.
   *
   * @var string[]
   */
  protected $iconsetLabels;

  /**
   * The theme handler service for Drupal.
   *
   * @var \Drupal\Core\Extension\ThemeHandlerInterface
   */
  protected $themeHandler;

  /**
   * Dependency injection class resolver.
   *
   * @var \Drupal\Core\DependencyInjection\ClassResolverInterface
   */
  protected $classResolver;

  /**
   * Create new instance of the Iconset plugin manager.
   *
   * @param Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The service class that manages the modules and invokes module hooks.
   * @param Drupal\Core\Extension\ThemeHandlerInterface $theme_handler
   *   The service class that manages themes.
   * @param Drupal\Core\Cache\CacheBackendInterface $cache_backend
   *   The caching backend to use for caching plugin definitions.
   * @param \Drupal\Core\DependencyInjection\ClassResolverInterface $class_resolver
   *   Dependency injection class resolver.
   */
  public function __construct(ModuleHandlerInterface $module_handler, ThemeHandlerInterface $theme_handler, CacheBackendInterface $cache_backend, ClassResolverInterface $class_resolver) {
    parent::__construct(
      'iconset',
      IconsetInterface::class,
      IconsetDefault::class,
      $module_handler,
      $cache_backend,
      'iconset'
    );

    $this->themeHandler = $theme_handler;
    $this->classResolver = $class_resolver;
  }

  /**
   * {@inheritdoc}
   */
  public function clearCachedDefinitions() {
    parent::clearCachedDefinitions();

    $this->iconsetLabels = NULL;
  }

  /**
   * {@inheritdoc}
   */
  protected function providerExists($provider) {
    // Allow both modules and themes to provide iconset definitions.
    return $this->extensionHandler->moduleExists($provider)
      || $this->themeHandler->themeExists($provider);
  }

  /**
   * {@inheritdoc}
   */
  protected function processDefinition(&$definition, $plugin_id) {
    parent::processDefinition($definition, $plugin_id);

    $provider = $definition['provider'];

    // @todo Explore the possibility of applying this type information to the
    // iconset definitions during the discovery with a custom YamlDiscovery.
    if ($provider === 'core') {
      $definition['type'] = 'core';
    }
    elseif ($this->extensionHandler->moduleExists($provider)) {
      $definition['type'] = 'module';
    }
    elseif ($this->themeHandler->themeExists($provider)) {
      $definition['type'] = 'theme';
    }
    else {
      unset($definition['type']);
    }
  }

  /**
   * Use the plugin manager discovery to find and process definitions.
   *
   * @return array
   *   An array of plugins discovered and processed by they discovery instance.
   */
  protected function findDefinitions() {
    $definitions = parent::findDefinitions();

    foreach ($definitions as $pluginId => $definition) {
      if (empty($definition['type'])) {
        unset($definitions[$pluginId]);
      }
    }

    return $definitions;
  }

  /**
   * {@inheritdoc}
   */
  public function getDiscovery() {
    if (!$this->discovery) {
      $discoveryDirs = $this->extensionHandler->getModuleDirectories()
        + $this->themeHandler->getThemeDirectories();

      $discovery = new YamlDiscovery('iconset', $discoveryDirs);
      $discovery->addTranslatableProperty('label', 'label_context');
      $discovery = new ContainerDerivativeDiscoveryDecorator($discovery);
      $this->discovery = $discovery;
    }
    return $this->discovery;
  }

  /**
   * {@inheritdoc}
   */
  public function getIconsets($include_internal = FALSE) {
    if (!isset($this->iconsetLabels)) {
      $this->iconsetLabels = [];

      foreach ($this->getDefinitions() as $definition) {
        $bucket = empty($definition['internal']) ? 'global' : 'internal';
        $this->iconsetLabels[$bucket][$definition['id']] = $definition['label'];
      }
    }

    $iconsets = $this->iconsetLabels['global'];
    if ($include_internal) {
      $iconsets += $this->iconsetLabels['internal'];
    }

    return $iconsets;
  }

  /**
   * {@inheritdoc}
   */
  public function initInstance(PluginInspectionInterface $iconset) {
    if (!$iconset instanceof IconsetInterface) {
      throw new \InvalidArgumentException('Iconsets need to implement \Drupal\iconset\IconsetInterface.');
    }

    $definition = $iconset->getPluginDefinition();

    // Instantiate a builder handler for the iconset if one is specified.
    if (!empty($definition['builder'])) {
      try {
        $builder = $this->classResolver->getInstanceFromDefinition($definition['builder']);
        $iconset->setBuilder($builder);
      }
      catch (\Exception $e) {
        // If class resolver throws an Exception it means the builder class is
        // not available or cannot be created. Log the error, but it is "safe"
        // to continue using the default builder.
        $this->getLogger('iconset')
          ->error('Unable to set Iconset builder %type due to @message: @backtrace_string', Error::decodeException($e));
      }
    }
  }

}
