<?php

namespace Drupal\iconset\Plugin\Field\FieldFormatter;

use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Field\FormatterBase;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\iconset\Exception\IconNotFoundException;
use Drupal\iconset\IconsetPluginManagerInterface;

/**
 * A Field formatter for displaying Iconset icons.
 *
 * @FieldFormatter(
 *   id = "iconset_icon_formatter",
 *   label = @Translation("Rendered icon"),
 *   field_types = {
 *     "iconset_icon",
 *   },
 * )
 */
class IconsetFieldFormatter extends FormatterBase implements ContainerFactoryPluginInterface {

  /**
   * The iconset manager.
   *
   * @var \Drupal\iconset\IconsetPluginManagerInteface
   */
  protected $iconsetManager;

  /**
   * Constructs an IconsetFieldFormatter object.
   *
   * @param array $configuration
   *   The field formatter configuration, including the field definition,
   *   formatter settings, view mode and third party settings.
   * @param string $plugin_id
   *   The plugin_id for the formatter.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\iconset\IconsetPluginManagerInterface $iconset_manager
   *   The iconset manager.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, IconsetPluginManagerInterface $iconset_manager) {
    parent::__construct(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['label'],
      $configuration['view_mode'],
      $configuration['third_party_settings']
    );

    $this->iconsetManager = $iconset_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('plugin.manager.iconset')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'width' => '',
      'height' => '',
      'css_class' => [],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $elements = parent::settingsForm($form, $form_state);
    $settings = $this->getSettings();

    $elements['width'] = [
      '#type' => 'number',
      '#title' => $this->t('Width (px)'),
      '#min' => 1,
      '#default_value' => $settings['width'] ?? NULL,
      '#description' => $this->t('The width attribute to set for this icon, will be overridden by CSS, and will auto size to maintain aspect ratio if height is set and this is left blank.'),
    ];

    $elements['height'] = [
      '#type' => 'number',
      '#title' => $this->t('Height (px)'),
      '#min' => 1,
      '#default_value' => $settings['height'] ?? NULL,
      '#description' => $this->t('The height attribute to set for this icon, will be overridden by CSS, and will auto size to maintain aspect ratio if width is set and this is left blank.'),
    ];

    $elements['css_class'] = [
      '#type' => 'css_class',
      '#title' => $this->t('Additional CSS classes'),
      '#default_value' => $settings['css_class'] ?? [],
    ];

    return $elements;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = parent::settingsSummary();
    $settings = $this->getSettings();

    if (!empty($settings['width'])) {
      if (!empty($settings['height'])) {
        $summary[] = $this->t('Icon size (pixels): %width x %height', [
          '%width' => $settings['width'],
          '%height' => $settings['height'],
        ]);
      }
      else {
        $summary[] = $this->t('Icon width: %width px', [
          '%width' => $settings['width'],
        ]);
      }
    }
    elseif (!empty($settings['height'])) {
      $summary[] = $this->t('Icon height: %height px', [
        '%height' => $settings['height'],
      ]);
    }

    if (!empty($settings['css_class'])) {
      $summary[] = $this->t('CSS Classes: %css_names', [
        '%css_names' => implode(' ', $settings['css_class']),
      ]);
    }

    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    $hasAltText = $this->fieldDefinition
      ->getFieldStorageDefinition()
      ->getSetting('alt_text');

    // Fetch the width and height settings, and prepare them to be applied
    // to the icon rendering options.
    $dimensions = array_filter([
      'width' => $this->getSetting('width'),
      'height' => $this->getSetting('height'),
    ]);

    $elements = [];
    foreach ($items as $delta => $item) {
      $options = $dimensions;

      if ($hasAltText && $item->alt) {
        $options['alt'] = $item->alt;
      }

      try {
        $iconset = $this->iconsetManager->createInstance($item->iconset);
        $build = $iconset->build($item->icon, $options);

        if ($this->getSetting('css_class')) {
          $build['#attributes']['class'] = isset($build['#attributes']['class'])
            ? array_merge($this->getSetting('css_class'), $build['#attributes']['class'])
            : $this->setting['css_class'];
        }

        $elements[$delta] = $build;
      }
      catch (PluginException | PluginNotFoundException | IconNotFoundException $e) {
        $elements[$delta] = [];
      }
    }

    return $elements;
  }

}
